; Copyright 2000, by Michael Vincent & Robin Kay. All rights reserved.
;
; This part of the source code is what really makes TSE 'tick'. Please only learn from this source, and not steal. Thank you.
;
;
.nolist    
#include "tse.inc"
.list

.org    saferam1-3
.db     $BB,$6D
 ret

; Task block details
flags_base          .equ 89F0h
flags_size          .equ 60
sp_size             .equ 2
task_sp             .equ 0
task_flags          .equ task_sp + sp_size
task_end            .equ task_flags + flags_size

; Block size
maxblocksize        .equ 100h

start:
 jr startkrnl
 jp verinfo    ; Library function 0, Get version/functions supported
 jp chkprog    ; Library function 1, Check program
 jp exitshell  ; Library function 2, Exit from shell
 jp taskswitch ; Library function 3, Switch task
 jp starttask  ; Library function 4, Start task
 jp endtask    ; Library function 5, End task
 jp forceyield ; Library function 6, Force yield

verinfo
 ld hl, %0000000001111111
majorver .equ 1
minorver .equ 6
 ld bc, (majorver*256)+minorver
 ret

startkrnl:

 ; Disable run indicator
 bcall(_runIndicOff)

 ; Preserve flags
 ld bc, flags_size
 ld hl, flags_base
 ld de, sysflags
 ldir

 ; Set active program
 ld hl, shellprog
 ld de, cprogram
 bcall(_mov9b)

 ; Look up the shell in the symbol table
 ld hl, cprogram
 rst 20h ;rMOV9TOOP1
 call starttask
 cp 0
 ret nz

 ; Check to see if enough free RAM for block copy buffer
 call chkEnoughForBuffer

 ; Preserve stack
 ld (sysstack), sp

 ; Exec
 jp taskenter
exitshell:

 ; Restore stack
 ld hl, (sysstack)
 ld sp, hl

 ; Copy out of active memory
 call cpy_prgm_out

 ; End task
 ld hl, shellprog
 rst 20h ;rMOV9TOOP1
 call endtask

 ; Clear screen
 bcall(_clrScrnFull)

 ; Display exit message
 ld hl, 0
 ld (currow), hl
 ld hl, msg
 bcall(_puts)

 ; Disable ON flag
 res onInterrupt, (iy+onFlags)

 ret

; Force the program to yield
;
; Inputs -
;  None
;
; Returns -
;  Nothing
;
forceyield:
 ld hl, shellprog
 rst 20h ;rMOV9TOOP1
 call taskswitch
 ret

; Load a program into active memory and run it
;
; Inputs -
;  OP1 should contain a program variable
;
; Returns -
;  Nothing
;  
taskswitch:

 ; Preserve sp
 ld (tempword), sp
 ld hl, (pptr_preserve)
 push hl
 ld de, (tempword)
 ld (hl), e
 inc hl
 ld (hl), d

 ; Preserve var
 ld hl, op1
 ld de, tprogram
 bcall(_mov9b)

 ; Preserve flags
 pop hl
 inc hl
 inc hl
 ex de, hl
 ld bc, flags_size
 ld hl, flags_base
 ldir

 ; Display message
 ld hl, 0
 ld (currow), hl
 set textInverse, (iy+textflags)
 ld hl, plswait
 bcall(_puts)

 ; Borrow system stack
 ld hl, (sysstack)
 ld sp, hl

 ; Check to see if enough RAM for block copy buffer
 call chkEnoughForBuffer

 ; Copy current program out of memory
 call cpy_prgm_out

 ; Copy program name to cprogram
 ld hl, tprogram
 ld de, cprogram
 bcall(_mov9b)

taskenter:

 ; Check to see if enough RAM for block copy buffer
 call chkEnoughForBuffer

 ; Load new program in
 call cpy_prgm_in

 ; Restore flags
 ld hl, (pptr_preserve)
 inc hl
 inc hl
 ld bc, flags_size
 ld de, flags_base
 ldir

 ; Restore sp
 ld hl, (pptr_preserve)
 bcall(_ldHLind)
 ld sp, hl

 ; Exec program
 ret

; Remove a task block from the end of a program variable and update its
; status.
;
; Inputs -
;  OP1 should contain a program variable
;
; Returns -
;  A will equal 0 on sucess otherwise it's an error code
endtask:

 ; Find the program and set PTR_'s
 call chkprog
 cp 0
 ret nz

 ; Check to see if task block doesn't exist
 ld hl, (fptr_prgtitle)
 ld a, (hl)
 cp ' '
 ld a, 4 ; Load error code 4 - Task block doesn't exist
 ret z

 ; Calculate task block size
 ld hl, (fptr_memreq)
 bcall(_ldHLind)
 ex de, hl
 ld hl, task_end
 add hl, de
 push hl
 push hl

 ; Find start of task block
 ex de, hl
 ld hl, (fptr_end)
 or a
 sbc hl, de

 ; Delete task block
 pop de
 bcall(_delmem)

 ; Update variable size field
 ld hl, (fptr_varsize)
 bcall(_ldHLind)
 pop bc
 or a
 sbc hl, bc
 ex de, hl
 ld hl, (fptr_varsize)
 ld (hl), e
 inc hl
 ld (hl), d

 ; Modify status byte and return
 ld hl, (fptr_prgtitle)
 ld a, ' '
 ld (hl), a

 ; Success
 xor a
 ret

; Add a task block onto the end of a program variable and update its
; status.
;
; Inputs -
;  OP1 should contain a program variable
;
; Returns -
;  A will equal 0 on sucess otherwise it's an error code
starttask:

 ; Find the program and set PTR_'s
 call chkprog
 cp 0
 ret nz

 ; Check to see if task block already exists
 ld hl, (fptr_prgtitle)
 ld a, (hl)
 cp '*'
 ld a, 4 ; Load error code 4 - Task block already exists
 ret z

 ; Check available memory
 ld hl, (fptr_memreq)
 bcall(_ldHLind)
 ld bc, task_end
 add hl, bc
 push hl
 push hl
 bcall(_enoughmem)
 ld a, 5 ; Load error code 5 - Insufficient memory
 ret c

 ; Insert memory
 ld de, (fptr_end)
 pop hl
 bcall(_insertmem)

 ; Update variable size field
 ld hl, (fptr_varsize)
 bcall(_ldHLind)
 pop bc
 add hl, bc
 ex de, hl
 ld hl, (fptr_varsize)
 ld (hl), e
 inc hl
 ld (hl), d

 ; Get address of taskblock
 ld hl, (fptr_memreq)
 bcall(_ldHLind)
 ld bc, (fptr_end)
 add hl, bc
 push hl
 push hl
 push hl

 ; Set start address
 ld de, userMem + 4
 ld hl, (fptr_prgtitle)
findcodeaddr:
 ld a, (hl)
 cp 0
 inc hl
 inc de
 jr nz, findcodeaddr
 inc de
 inc de
 pop hl
 dec hl
 ld (hl), d
 dec hl
 ld (hl), e

 ; Set sp
 ld hl, (fptr_varsize)
 bcall(_ldHLind)
 ld bc, userMem - (3 + task_end + 2)
 add hl, bc
 ex de, hl
 pop hl
 ld (hl), e
 inc hl
 ld (hl), d

 ; Save flags into taskblock
 pop hl
 inc hl
 inc hl
 ex de, hl
 ld bc, flags_size
 ld hl, sysflags
 ldir

 ; Update status bytes
 ld hl, (fptr_prgtitle)
 ld a, '*'
 ld (hl), a

 ; Success
 xor a
 ret
 
; Locate a TSE program in memory and set pointers.
;
; Inputs -
;  OP1 should contain a program variable
;
; Returns -
;  A will equal 0 on sucess otherwise it's an error code
;  HL points to prgtitle
;  (fptr_varsize), (fptr_code) and (fptr_end) set point to program
;
chkprog:

 ; Check to see if variable exists
 bcall(_ChkFindSym)
 ex de, hl
 ld a, 1 ; Load error code 1 - Variable does not exist
 ret c ; Return if error

 ; Check to see if variable is in RAM
 ld a, b
 cp 0
 ;OLD: ld a, 2 ; Load error code 2 - Variable is stored in FLASH-ROM
 ;OLD: ret nz ; Return if error
 jr z, notFlash
 ld de, saferam5
 ld bc, 256
 bcall(_FlashToRam)
 ld hl, saferam5 + 9
 ld c, (hl)
 ld b, 0
 inc bc
 add hl, bc
 call notFlash
 cp 0
 ret nz
 ld a, 2 ; Load error code 2 - Variable is stored in FLASH-ROM
 ret
notFlash:

 ; Get pointer to "variable size field"
 ld (fptr_varsize), hl

 ; Get pointer to "end of variable"
 push hl
 bcall(_ldHLind)
 ld de, (fptr_varsize)
 inc de
 inc de
 add hl, de
 ld (fptr_end), hl
 pop hl

 ; Check that the TSE header is valid
 ld b, 4
 ld de, 5
 add hl, de
 ld de, tsevalid
cmpStr:
 ld a, (de)
 ld c, (hl)
 cp c ; Compare strings
 ld a, 3 ; Load error code 3 - Not a valid TSE program
 ret nz ; Return if error
 inc hl ; Next byte of header
 inc de ; Next byte of tsevalid
 djnz cmpStr

 ; Get pointer to "program title"
 ld (fptr_prgtitle), hl

 ; Get pointer to "mem required"
 call skipstr
 ld (fptr_memreq), hl

 ; Load HL with program title
 ld hl, (fptr_prgtitle)

 ; All done
 xor a ; Load error code 0 - Sucess
 ret

; Moves the active program into a variable
;
; Inputs -
;  (cprogram) should contain the name of the program to load in
;
; Returns -
;  Program is moved into variable
;
cpy_prgm_out:

 ; Load prgm name into OP1
 ld hl, cprogram
 rst 20h ;rMOV9TOOP1

 ; Get src
 ld hl, userMem
 ld (cpysrc), hl

 ; Clear tempword
 ld hl, 5
 ld (tempword), hl

 ; Get program size
 ld hl, (progsize)
 ld (bytes2go), hl

 ; Update size field of var
 push hl
 bcall(_chkfindsym)
 ex de, hl
 pop de
 inc de
 inc de
 inc de
 ld (hl), e
 inc hl
 ld (hl), d

cpyout_loop:

 ; Get copy block size
 call calcblock

 ; Calculate cpydest
 bcall(_chkfindsym)
 ex de, hl
 ld bc, (tempword)
 add hl, bc
 ex de, hl

 ; Increase size of dest
 ld hl, (blocksize)
 bcall(_insertmem)

 ; Copy data
 ld hl, (cpysrc)
 ld bc, (blocksize)
 ldir

 ; Increase tempword
 ld hl, (tempword)
 ld bc, (blocksize)
 add hl, bc
 ld (tempword), hl

 ; Decrease size of src
 ld hl, (cpysrc)
 ld de, (blocksize)
 bcall(_delmem) 

 ; Loop if not finished
 ld hl, (bytes2go)
 xor a
 cp h
 jr nz, cpyout_loop
 cp l
 jr nz, cpyout_loop

 ret

; Moves the selected variable into the active program
;
; Inputs -
;  (cprogram) should contain the name of the program to load in
;
; Returns -
;  Program is moved into variable
;
cpy_prgm_in:

 ; Load prgm name into OP1
 ld hl, cprogram
 rst 20h ;rMOV9TOOP1

 ; Get program size
 bcall(_chkfindsym)
 ex de, hl
 push hl
 bcall(_ldhlind)
 dec hl
 dec hl
 dec hl
 ld (bytes2go), hl
 ld (progsize), hl

 ; Update size field of var
 pop hl
 ld de, 3
 ld (hl), e
 inc hl
 ld (hl), d

 ; Get dest
 ld hl, userMem
 ld (cpydest), hl

cpyin_loop:

 ; Get copy block size
 call calcblock

 ; Increase size of dest
 ld hl, (blocksize)
 ld de, (cpydest)
 bcall(_insertmem)

 ; Calculate cpysrc
 bcall(_chkfindsym)
 ex de, hl
 ld bc, 5
 add hl, bc
 push hl

 ; Copy data
 ld bc, (blocksize)
 ld de, (cpydest)
 ldir
 ld (cpydest), de

 ; Decrease size of src
 pop hl
 ld de, (blocksize)
 bcall(_delmem) 

 ; Loop if not finished
 ld hl, (bytes2go)
 xor a
 cp h
 jr nz, cpyin_loop
 cp l
 jr nz, cpyin_loop

 ; Set pptr_code
 ld hl, userMem + 4
 call skipstr
 inc hl
 inc hl
 ld (pptr_code), hl

 ; Set pptr_preserve
 ld hl, (progsize)
 ld bc, userMem - task_end
 add hl, bc
 ld (pptr_preserve), hl

 ret

; Copy block size calulating routine
;
calcblock:
 ld bc, maxblocksize
smaller:
 dec bc
 ld hl, (bytes2go)
 or a
 sbc hl, bc
 jr c, smaller
 ld (bytes2go), hl
 ld (blocksize), bc
 ret

; Skip string routine
;
skipstr:
 ld a, (hl)
 cp 0
 inc hl
 jr nz, skipstr
 ret

; Check to see if enough free RAM for block copy buffer
chkEnoughForBuffer:
 ld hl, maxblocksize
 bcall(_EnoughMem)
 ret nc
 pop hl ; Get return address off stack
 ret

; Please Wait
;
plswait:
.db "<WAIT>",0

; Exit Message
;
msg:
.db "TSE v1.6",0

; The 4 byte TSE header for use by the validation code
;
tsevalid:
.db "TSE", 1

; Name of program to load as shell
;
shellprog:
.db ProtProgObj, "UTOPIA",0,0

; Memory equates
;
data:

; Memory equates (Temporary)
bytes2go          .equ saferam2 + 400
blocksize         .equ bytes2go + 2
cpysrc            .equ blocksize + 2
cpydest           .equ cpysrc + 2
tempword          .equ cpydest + 2
tprogram          .equ tempword + 2

.end
